from toolmate import showErrors, showRisk, executeToolFunction, getPythonFunctionResponse, getPygmentsStyle, fineTunePythonCode, confirmExecution, useChatSystemMessage
from toolmate import config
from toolmate import print1, print2, print3, validParameters, getGoogleGenAIClient
import re, traceback, openai, pprint, copy, textwrap, json, pygments, codecs
from pygments.lexers.python import PythonLexer
from prompt_toolkit import print_formatted_text, HTML
from prompt_toolkit.formatted_text import PygmentsTokens
from prompt_toolkit import prompt
from typing import Optional


def check_errors(func):
    def wrapper(*args, **kwargs):
        def finishError():
            config.stopSpinning()
            return "[INVALID]"
        try:
            return func(*args, **kwargs)
        except:
            print(traceback.format_exc())
            return finishError()
    return wrapper

@check_errors
def checkCompletion():
    getGoogleGenAIClient().chat.completions.create(
        model="gemini-1.5-flash-8b",
        messages=[{"role": "user", "content" : "hello"}],
        n=1,
        max_tokens=10,
    )

@check_errors
def riskAssessment(code):
    content = f"""You are a senior python engineer.
Assess the risk level of damaging my device upon executing the python code that I will provide for you.
Answer me either 'high', 'medium' or 'low', without giving me any extra information.
e.g. file deletions or similar significant impacts are regarded as 'high' level.
Acess the risk level of this Python code:
```
{code}
```"""
    try:
        answer = getSingleChatResponse(content, temperature=0.0, keepSystemMessage=True)
        if not answer:
            answer = "high"
        answer = re.sub("[^A-Za-z]", "", answer).lower()
        if not answer in ("high", "medium", "low"):
            answer = "high"
        return answer
    except:
        return "high"

@check_errors
def autoCorrectPythonCode(code, trace):
    for i in range(config.max_consecutive_auto_correction):
        userInput = f"Original python code:\n```\n{code}\n```\n\nTraceback:\n```\n{trace}\n```"
        messages = [{"role": "user", "content" : userInput}]
        print3(f"Auto-correction attempt: {(i + 1)}")
        function_call_message, function_call_response = CallGoogleAI.getSingleFunctionCallResponse(messages, "correct_python_code")
        # display response
        print1(config.divider)
        if config.developer:
            print(function_call_response)
        else:
            print1("Executed!" if function_call_response == "EXECUTED" else "Failed!")
        if function_call_response == "EXECUTED":
            break
        else:
            code = json.loads(function_call_message["function_call"]["arguments"]).get("code")
            trace = function_call_response
        print1(config.divider)
    # return information if any
    if function_call_response == "EXECUTED":
        pythonFunctionResponse = getPythonFunctionResponse(code)
        if pythonFunctionResponse:
            return json.dumps({"information": pythonFunctionResponse})
        else:
            return ""
    # ask if user want to manually edit the code
    print1(f"Failed to execute the code {(config.max_consecutive_auto_correction + 1)} times in a row!")
    print1("Do you want to manually edit it? [y]es / [N]o")
    confirmation = prompt(style=config.promptStyle2, default="N")
    if confirmation.lower() in ("y", "yes"):
        config.defaultEntry = f"```python\n{code}\n```"
        return ""
    else:
        return "[INVALID]"

def convertFunctionSignaturesIntoTools(functionSignatures):
    return [{"type": "function", "function": functionSignature} for functionSignature in functionSignatures]

def getToolArgumentsFromStreams(completion):
    toolArguments = {}
    for event in completion:
        delta = event.choices[0].delta
        if delta and delta.tool_calls:
            for tool_call in delta.tool_calls:
                # handle functions
                if tool_call.function:
                    func_index = tool_call.index
                    if func_index in toolArguments:
                        toolArguments[func_index] += tool_call.function.arguments
                    else:
                        toolArguments[func_index] = tool_call.function.arguments
                # may support non functions later
    return toolArguments

@check_errors
def getSingleChatResponse(userInput, messages=[], temperature=None, prefill: Optional[str]=None, stop: Optional[list]=None, keepSystemMessage: bool=False):
    """
    non-streaming single call
    """
    if userInput:
        item = {"role": "user", "content" : userInput}
        if messages and messages[-1].get("role", "") == "assistant":
            messages.insert(-1, item)
        else:
            messages.append(item)
    chatMessages = copy.deepcopy(messages) if keepSystemMessage else useChatSystemMessage(copy.deepcopy(messages))
    if prefill is not None:
            chatMessages.append({'role': 'assistant', 'content': prefill})
    try:
        completion = getGoogleGenAIClient().chat.completions.create(
            model=config.googleaiApi_tool_model,
            messages=chatMessages,
            n=1,
            temperature=temperature if temperature is not None else config.llmTemperature,
            max_tokens=config.googleaiApi_tool_model_max_tokens,
            stop=stop if stop else None,
        )
        return completion.choices[0].message.content
    except:
        return ""

def runSingleFunctionCall(messages, function_name):
    messagesCopy = copy.deepcopy(messages)
    try:
        _, function_call_response = getSingleFunctionCallResponse(messages, function_name)
        function_call_response = function_call_response if function_call_response else config.toolTextOutput
        messages[-1]["content"] += f"""\n\nAvailable information:\n{function_call_response}"""
        config.toolTextOutput = ""
    except:
        showErrors()
        return messagesCopy
    return messages

@check_errors
def getSingleFunctionCallResponse(messages: list[dict], function_name: str, temperature=None, **kwargs):
    functionSignatures = [config.toolFunctionSchemas[function_name]]
    completion = getGoogleGenAIClient().chat.completions.create(
        model=config.googleaiApi_tool_model,
        messages=messages,
        n=1,
        temperature=temperature if temperature is not None else config.llmTemperature,
        max_tokens=config.googleaiApi_tool_model_max_tokens,
        tools=convertFunctionSignaturesIntoTools(functionSignatures),
        tool_choice={"type": "function", "function": {"name": function_name}},
        stream=False,
        **kwargs,
    )
    function_call_message = completion.choices[0].message
    tool_call = function_call_message.tool_calls[0]
    func_arguments = tool_call.function.arguments
    codecs.decode(func_arguments, "unicode_escape")
    function_call_message_mini = {
        "role": "assistant",
        "content": "",
        "function_call": {
            "name": tool_call.function.name,
            "arguments": func_arguments,
        }
    }
    function_call_response = CallGoogleAI.finetuneSingleFunctionCallResponse(func_arguments, function_name)
    return function_call_message_mini, function_call_response

def finetuneSingleFunctionCallResponse(func_arguments, function_name):
    def notifyDeveloper(func_name):
        if config.developer:
            #print1(f"running tool '{func_name}' ...")
            print_formatted_text(HTML(f"<{config.terminalPromptIndicatorColor2}>Running tool</{config.terminalPromptIndicatorColor2}> <{config.terminalCommandEntryColor2}>'{func_name}'</{config.terminalCommandEntryColor2}> <{config.terminalPromptIndicatorColor2}>...</{config.terminalPromptIndicatorColor2}>"))
    # ChatGPT's built-in function named "python"
    if function_name == "python":
        notifyDeveloper(function_name)
        python_code = textwrap.dedent(func_arguments)
        refinedCode = fineTunePythonCode(python_code)

        print1(config.divider)
        print2("running python code ...")
        risk = riskAssessment(python_code)
        showRisk(risk)
        if config.developer or config.codeDisplay:
            print("```")
            #print(python_code)
            # pygments python style
            tokens = list(pygments.lex(python_code, lexer=PythonLexer()))
            print_formatted_text(PygmentsTokens(tokens), style=getPygmentsStyle())
            print("```")
        print1(config.divider)

        config.stopSpinning()
        if not config.runPython:
            info = {"information": python_code}
            return json.dumps(info)
        elif confirmExecution(risk):
            print1("Do you want to continue? [y]es / [N]o")
            confirmation = prompt(style=config.promptStyle2, default="y")
            if not confirmation.lower() in ("y", "yes"):
                info = {"information": python_code}
                return json.dumps(info)
        try:
            exec(refinedCode, globals())
            function_response = getPythonFunctionResponse(refinedCode)
        except:
            trace = showErrors()
            print1(config.divider)
            if config.max_consecutive_auto_correction > 0:
                return autoCorrectPythonCode(refinedCode, trace)
            else:
                return "[INVALID]"
        if function_response:
            info = {"information": function_response}
            function_response = json.dumps(info)
    # known unwanted functions are handled here
    elif function_name in ("translate_text",):
        # "translate_text" has two arguments, "text", "target_language"
        # handle known and unwanted function
        function_response = "[INVALID]" 
    # handle unexpected function
    elif not function_name in config.toolFunctionMethods:
        if config.developer:
            print1(f"Unexpected function: {function_name}")
            print1(config.divider)
            print(func_arguments)
            print1(config.divider)
        function_response = "[INVALID]"
    else:
        notifyDeveloper(function_name)
        fuction_to_call = config.toolFunctionMethods[function_name]
        # convert the arguments from json into a dict
        function_args = json.loads(func_arguments)
        function_response = fuction_to_call(function_args)
    return function_response


class CallGoogleAI:

    @staticmethod
    @check_errors
    def checkCompletion():
        checkCompletion()

    @staticmethod
    @check_errors
    def autoCorrectPythonCode(code, trace):
        return autoCorrectPythonCode(code, trace)

    @staticmethod
    @check_errors
    def getSingleChatResponse(userInput, messages=[], temperature=None, prefill: Optional[str]=None, stop: Optional[list]=None, keepSystemMessage: bool=False):
        return getSingleChatResponse(userInput, messages, temperature, prefill, stop, keepSystemMessage)

    @staticmethod
    def finetuneSingleFunctionCallResponse(func_arguments, function_name):
        return finetuneSingleFunctionCallResponse(func_arguments, function_name)

    @staticmethod
    def runSingleFunctionCall(messages, function_name):
        return runSingleFunctionCall(messages, function_name)

    @staticmethod
    @check_errors
    def getSingleFunctionCallResponse(messages: list[dict], function_name: str, temperature=None, **kwargs):
        return getSingleFunctionCallResponse(messages, function_name, temperature, **kwargs)

    @staticmethod
    @check_errors
    def regularCall(messages: dict, **kwargs):
        chatMessages = useChatSystemMessage(copy.deepcopy(messages))
        return getGoogleGenAIClient().chat.completions.create(
            model=config.googleaiApi_tool_model,
            messages=chatMessages,
            n=1,
            temperature=config.llmTemperature,
            max_tokens=config.googleaiApi_tool_model_max_tokens,
            stream=True,
            **kwargs,
        )

    @staticmethod
    @check_errors
    @check_errors
    def getDictionaryOutput(messages: list, schema: dict, **kwargs) -> dict:
        completion = getGoogleGenAIClient().chat.completions.create(
            model=config.googleaiApi_tool_model,
            messages=messages,
            n=1,
            temperature=config.llmTemperature,
            max_tokens=config.googleaiApi_tool_model_max_tokens,
            tools=[{"type": "function", "function": schema}],
            tool_choice={"type": "function", "function": {"name": schema["name"]}},
            stream=False,
            **kwargs,
        )
        outputMessage = completion.choices[0].message
        if hasattr(outputMessage, "tool_calls") and outputMessage.tool_calls:
            function_arguments = outputMessage.tool_calls[0].function.arguments
            responseDict = json.loads(codecs.decode(function_arguments, "unicode_escape"))
        else:
            #print("Failed to output structered data!")
            if hasattr(outputMessage, "content") and outputMessage.content:
                return codecs.decode(outputMessage.content, "unicode_escape")
            return {}
        return responseDict

    # Auto Function Call equivalence

    @staticmethod
    def runToolCall(messages: dict):
        if not config.selectedTool:
            return CallGoogleAI.regularCall(messages)
        else:
            # 2. Tool Selection
            if config.selectedTool and not config.selectedTool == "chat" and config.selectedTool in config.toolFunctionSchemas:
                tool_name = config.selectedTool
                tool_schema = config.toolFunctionSchemas[tool_name]
                config.selectedTool = ""
            else:
                return CallGoogleAI.regularCall(messages)
            # 3. Parameter Extraction
            if config.developer:
                print1("extracting parameters ...")
            try:
                if not tool_schema["parameters"]["properties"]:
                    # Execute function directly
                    tool_parameters = {}
                    tool_response = executeToolFunction(func_arguments=tool_parameters, function_name=tool_name)
                else:
                    tool_parameters = CallGoogleAI.getDictionaryOutput(messages=messages, schema=tool_schema)
                    if isinstance(tool_parameters, str):
                        tool_response = tool_parameters
                    elif not validParameters(tool_parameters, tool_schema["parameters"]["required"]):
                        return CallGoogleAI.regularCall(messages)
                    else:
                        # 4. Function Execution
                        tool_response = executeToolFunction(func_arguments=tool_parameters, function_name=tool_name)
            except:
                print(traceback.format_exc())
                tool_response = "[INVALID]"
            # 5. Chat Extension
            if tool_response == "[INVALID]":
                # invalid tool call; return a regular call instead
                return CallGoogleAI.regularCall(messages)
            else:
                # record tool selection
                #config.currentMessages[-1]["tool"] = tool_name
                if tool_response:
                    if config.developer:
                        print2(config.divider)
                        print2("Tool output:")
                        print(tool_response)
                        print2(config.divider)
                    # update message chain
                    user_request = messages[-1]["content"]
                    messages[-1]["content"] = f"""# Provided Context

{tool_response}

# My question:

{user_request}

# Instruction

Select all the relevant information from the provided context to answer my question."""
                    return CallGoogleAI.regularCall(messages)
                elif (not config.currentMessages[-1].get("role", "") == "assistant" and not config.currentMessages[-2].get("role", "") == "assistant") or (config.currentMessages[-1].get("role", "") == "system" and not config.currentMessages[-2].get("role", "") == "assistant"):
                    # tool function executed without chat extension
                    config.currentMessages.append({"role": "assistant", "content": config.toolTextOutput if config.toolTextOutput else "Done!"})
                    config.toolTextOutput = ""
                    config.conversationStarted = True
                    return None

    @staticmethod
    def extractToolParameters(schema: dict, ongoingMessages: list = [], **kwargs) -> dict:
        """
        Extract action parameters
        """
        parameters = CallGoogleAI.getDictionaryOutput(messages=ongoingMessages, schema=schema, **kwargs)
        if config.developer:
            print2("```parameters")
            pprint.pprint(parameters)
            print2("```")
        return parameters



import sys, traceback, openai, os, json, traceback, re, textwrap
from PySide6.QtCore import QRunnable, Slot, Signal, QObject, QThreadPool
from toolmate import config
from toolmate.utils.tts_utils import TTSUtil

class WorkerSignals(QObject):
    '''
    Defines the signals available from a running worker thread.

    Supported signals are:

    finished
        No data

    error
        tuple (exctype, value, traceback.format_exc() )

    result
        object data returned from processing, anything

    progress
        int indicating % progress

    '''
    finished = Signal()
    error = Signal(tuple)
    result = Signal(object)
    progress = Signal(str)


class Worker(QRunnable):
    '''
    Worker thread

    Inherits from QRunnable to handler worker thread setup, signals and wrap-up.

    :param callback: The function callback to run on this worker thread. Supplied args and
                     kwargs will be passed through to the runner.
    :type callback: function
    :param args: Arguments to pass to the callback function
    :param kwargs: Keywords to pass to the callback function

    '''

    def __init__(self, fn, *args, **kwargs):
        super(Worker, self).__init__()

        # Store constructor arguments (re-used for processing)
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.signals = WorkerSignals()

        # Add the callback to our kwargs
        self.kwargs["progress_callback"] = self.signals.progress

    @Slot()
    def run(self):
        '''
        Initialise the runner function with passed args, kwargs.
        '''

        # assign a reference to this current thread
        #config.workerThread = QThread.currentThread()

        # Retrieve args/kwargs here; and fire processing using them
        try:
            result = self.fn(*self.args, **self.kwargs)
        except:
            traceback.print_exc()
            exctype, value = sys.exc_info()[:2]
            self.signals.error.emit((exctype, value, traceback.format_exc()))
        else:
            self.signals.result.emit(result)  # Return the result of the processing
        finally:
            self.signals.finished.emit()  # Done

class QtResponseStreamer:

    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.threadpool = QThreadPool()

    def readAnswer(self, answer):
        # read the chunk when there is a punctuation
        #if answer in string.punctuation and config.llmTextChunk:
        if re.search(config.tts_startReadPattern, answer) and config.llmTextChunk:
            # read words when there a punctuation
            chunk = config.llmTextChunk + answer
            # reset config.llmTextChunk
            config.llmTextChunk = ""
            # play with tts
            if config.ttsOutput:
                TTSUtil.play(re.sub(config.tts_doNotReadPattern, "", chunk))
        else:
            # append to a chunk for reading
            config.llmTextChunk += answer

    def processCompletion(self, completion, progress_callback, openai: bool):
        #openai = True if config.llmInterface in ("chatgpt", "letmedoit", "groq", "llamacppserver") else False
        config.new_chat_response = ""

        def finishOutputs(chat_response):
            # reset config.llmTextChunk
            config.llmTextChunk = ""
            # add chat response to messages
            if chat_response:
                config.new_chat_response = chat_response
            if hasattr(config, "currentMessages") and chat_response:
                config.currentMessages.append({"role": "assistant", "content": chat_response})
            # finishing
            if hasattr(config, "conversationStarted"):
                config.conversationStarted = True
            self.streaming_finished = True

        chat_response = ""
        firstEvent = True
        self.streaming_finished = False
        for event in completion:
            # check if users manually stop
            stop_file = ".stop_output"
            if os.path.isfile(stop_file):
                self.streaming_finished = True
                os.remove(stop_file)
            if not self.streaming_finished:
                # RETRIEVE THE TEXT FROM THE RESPONSE
                if openai:
                    # openai
                    # when open api key is invalid for some reasons, event response in string
                    answer = event if isinstance(event, str) else event.choices[0].delta.content
                elif isinstance(event, dict):
                    if "message" in event:
                        # ollama chat
                        answer = event["message"].get("content", "")
                    else:
                        # llama.cpp chat
                        answer = event["choices"][0]["delta"].get("content", "")
                else:
                    # vertex ai
                    answer = event.text
                # transform
                if hasattr(config, "outputTransformers"):
                    for transformer in config.outputTransformers:
                        answer = transformer(answer)
                # STREAM THE ANSWER
                if answer is not None:
                    if firstEvent:
                        firstEvent = False
                        answer = answer.lstrip()
                    # display the chunk
                    chat_response += answer
                    # display text chunk
                    progress_callback.emit(answer)
                    # speak streaming words
                    self.readAnswer(answer)
            else:
                if hasattr(config, "llamacppserver_chat_client") and config.llamacppserver_chat_client is not None:
                    config.llamacppserver_chat_client.close()
                    config.llamacppserver_chat_client = None
                elif hasattr(config, "llamacppserver_tool_client") and config.llamacppserver_tool_client is not None:
                    config.llamacppserver_tool_client.close()
                    config.llamacppserver_tool_client = None
                finishOutputs(chat_response)
                return None
        
        if config.ttsOutput and config.llmTextChunk:
            TTSUtil.play(config.llmTextChunk)
            config.llmTextChunk = ""
        finishOutputs(chat_response)

    def workOnCompletion(self, completion, openai: bool):
        # Pass the function to execute
        worker = Worker(self.processCompletion, completion, openai=openai) # Any other args, kwargs are passed to the run function
        worker.signals.result.connect(self.parent.processResponse) # process the output return by self.parent.getResponse
        worker.signals.progress.connect(self.parent.streamResponse)
        # Connection
        #worker.signals.finished.connect(None)
        # Execute
        self.threadpool.start(worker)


"""
ToolMate AI Plugin - analyze files

analyze files with integrated "AutoGen Retriever"

[TOOL_CALL]
"""

if not config.isLite:

    from toolmate import config, is_valid_image_file, ragRefineDocsPath, ragGetSplits, ragSearchContext, getRagPrompt
    from toolmate import print2, print3
    from toolmate.utils.call_llm import CallLLM
    from toolmate.autoretriever import AutoGenRetriever
    import os
    #from PIL import Image


    def analyze_files(function_args):
        query = function_args.get("query") # required
        files = function_args.get("filepath") # required
        if os.path.exists(files):
            if os.path.isfile(files) and is_valid_image_file(files):
                # call function "analyze image" instead if it is an image
                function_args = {
                    "query": query,
                    "files": [files],
                }
                print3("Running tool: 'analyze_images'")
                return config.toolFunctionMethods["analyze_images"](function_args)
            config.stopSpinning()
            if config.rag_useAutoRetriever and not config.llmInterface in ("vertexai", "googleai"):
                print2("AutoGen Retriever launched!")
                last_message = AutoGenRetriever().getResponse(files, query, True)
                config.currentMessages += last_message
                print2("AutoGen Retriever closed!")
            else:
                print2("Retriever utility launched!")

                retrievedContext = ragSearchContext(ragGetSplits(ragRefineDocsPath(files)), query)

                formatted_prompt = getRagPrompt(query, retrievedContext)

                messages = config.currentMessages[:-1] + [{"role": "user", "content" : formatted_prompt}]

                completion = CallLLM.regularCall(messages)
                print2(config.divider)
                config.toolmate.streamCompletion(completion)
                print2(config.divider)

                print2("Retriever utility closed!")
            return ""

        return "[INVALID]"

    functionSignature = {
        "examples": [
            "analyze files",
            "retrieve file information",
        ],
        "name": "analyze_files",
        "description": "Retrieve information from files",
        "parameters": {
            "type": "object",
            "properties": {
                "query": {
                    "type": "string",
                    "description": "Detailed queries about the given files",
                },
                "filepath": {
                    "type": "string",
                    "description": """Return a directory or non-image file path. Return an empty string '' if it is not given.""",
                },
            },
            "required": ["query", "filepath"],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=analyze_files)


"""
ToolMate AI Plugin - analyze images

analyze images

Platform: llamacpp, ollama
Model: llava <- customizable
To customise:
Change in config.py:
llamacppVisionModel_model_path
llamacppVisionModel_clip_model_path
ollamaVisionModel

Platform: gemini
Model: Gemini Pro Vision

Platform: chaptgpt, letmedoit
Model "gpt-4o"
Reference: https://platform.openai.com/docs/guides/vision

[TOOL_CALL]
"""

if not config.isLite:

    from toolmate import config, print1, print2, is_valid_image_file, is_valid_image_url, startLlamacppVisionServer, stopLlamacppVisionServer, is_valid_url, encode_image, runToolMateCommand, getLlamacppServerClient
    from toolmate.utils.call_chatgpt import check_openai_errors
    import os
    from openai import OpenAI
    from toolmate.geminiprovision import GeminiProVision
    from toolmate.utils.call_ollama import CallOllama

    @check_openai_errors
    def analyze_images_llamacpp(function_args):
        from toolmate import config

        llmInterface = "llamacpp"

        if llmInterface == "vertexai":
            answer = GeminiProVision(temperature=config.llmTemperature).analyze_images_llamacpp(function_args)
            if answer:
                config.toolTextOutput = answer
                return ""
            else:
                return "[INVALID]"
        elif llmInterface in ("chatgpt", "letmedoit") and not config.openaiApiKey:
            return "OpenAI API key not found!"

        query = function_args.get("query") # required
        files = function_args.get("image_filepath") # required
        #print(files)
        if isinstance(files, str):
            if not files.startswith("["):
                files = f'["{files}"]'
            files = eval(files)

        filesCopy = files[:]
        for item in filesCopy:
            if os.path.isdir(item):
                for root, _, allfiles in os.walk(item):
                    for file in allfiles:
                        file_path = os.path.join(root, file)
                        files.append(file_path)
                files.remove(item)

        content = []
        # valid image paths
        for i in files:
            if is_valid_url(i) and is_valid_image_url(i):
                content.append({"type": "image_url", "image_url": {"url": i,},})
            elif os.path.isfile(i) and is_valid_image_file(i):
                content.append({"type": "image_url", "image_url": {"url": encode_image(i)},})
            else:
                files.remove(i)

        if content:
            # start llama.cpp vision server
            startLlamacppVisionServer()

            content.insert(0, {"type": "text", "text": query,})

            response = client.chat.completions.create(
                model="gpt-4o",
                messages=[
                    {
                    "role": "user",
                    "content": content,
                    }
                ],
                max_tokens=4096,
            )
            answer = response.choices[0].message.content
            config.toolTextOutput = answer

            # display answer
            print2("```assistant")
            print1(answer)
            print2("```")

            # stop llama.cpp vision server
            if llmInterface == "llamacpp":
                stopLlamacppVisionServer()

            return ""
        return "[INVALID]"

    functionSignature = {
        "examples": [
            "describe image",
            "compare images",
            "analyze image",
        ],
        "name": "analyze_images_llamacpp",
        "description": "Describe or compare images with Llama.cpp",
        "parameters": {
            "type": "object",
            "properties": {
                "query": {
                    "type": "string",
                    "description": "Questions or requests that users ask about the given images",
                },
                "image_filepath": {
                    "type": "string",
                    "description": """Return a list of image paths or urls, e.g. '["image1.png", "/tmp/image2.png", "https://letmedoit.ai/image.png"]'. Return '[]' if image path is not provided.""",
                },
            },
            "required": ["query", "image_filepath"],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=analyze_images_llamacpp)



import vertexai, os, traceback, argparse
from vertexai.generative_models import GenerativeModel, Content, Part
from vertexai.generative_models._generative_models import (
    GenerationConfig,
    HarmCategory,
    HarmBlockThreshold,
)
from toolmate import config
from toolmate import print1, print2, print3, toggleinputaudio, toggleoutputaudio
from toolmate.utils.streaming_word_wrapper import StreamingWordWrapper
from toolmate.utils.single_prompt import SinglePrompt
from toolmate.utils.tool_plugins import Plugins

from prompt_toolkit.completion import WordCompleter, FuzzyCompleter
from prompt_toolkit.styles import Style
from prompt_toolkit import PromptSession
from prompt_toolkit.history import FileHistory
from prompt_toolkit.shortcuts import clear
from pathlib import Path
import threading


# Install google-cloud-aiplatform FIRST!
#!pip install --upgrade google-cloud-aiplatform


class GeminiPro:

    def __init__(self, name="Gemini Pro", temperature=0.9, max_output_tokens=8192):
        # authentication
        if os.environ["GOOGLE_APPLICATION_CREDENTIALS"] and "Vertex AI" in config.enabledGoogleAPIs:
            self.runnable = True
        else:
            print("Vertex AI is not enabled!")
            print("Read https://github.com/eliranwong/toolmate/blob/main/package/toolmate/docs/Google%20Cloud%20Service%20Credential%20Setup.md for setting up Google API.")
            self.runnable = False
        # initiation
        vertexai.init()
        self.name, self.temperature = name, temperature
        self.generation_config=GenerationConfig(
            temperature=temperature, # 0.0-1.0; default 0.9
            max_output_tokens=max_output_tokens, # default
            candidate_count=1,
        )
        # Note: BLOCK_NONE is not allowed
        self.safety_settings={
            HarmCategory.HARM_CATEGORY_UNSPECIFIED: HarmBlockThreshold.BLOCK_ONLY_HIGH,
            HarmCategory.HARM_CATEGORY_DANGEROUS_CONTENT: HarmBlockThreshold.BLOCK_ONLY_HIGH,
            HarmCategory.HARM_CATEGORY_HATE_SPEECH: HarmBlockThreshold.BLOCK_ONLY_HIGH,
            HarmCategory.HARM_CATEGORY_HARASSMENT: HarmBlockThreshold.BLOCK_ONLY_HIGH,
            HarmCategory.HARM_CATEGORY_SEXUALLY_EXPLICIT: HarmBlockThreshold.BLOCK_ONLY_HIGH,
        }
        """
        Alternately,
        from google.cloud.aiplatform_v1beta1.types.content import SafetySetting
        self.safety_settings = [
            SafetySetting(category=HarmCategory.HARM_CATEGORY_UNSPECIFIED, threshold=HarmBlockThreshold.BLOCK_ONLY_HIGH)
            SafetySetting(category=HarmCategory.HARM_CATEGORY_DANGEROUS_CONTENT, threshold=HarmBlockThreshold.BLOCK_ONLY_HIGH)
            SafetySetting(category=HarmCategory.HARM_CATEGORY_HATE_SPEECH, threshold=HarmBlockThreshold.BLOCK_ONLY_HIGH)
            SafetySetting(category=HarmCategory.HARM_CATEGORY_HARASSMENT, threshold=HarmBlockThreshold.BLOCK_ONLY_HIGH)
            SafetySetting(category=HarmCategory.HARM_CATEGORY_SEXUALLY_EXPLICIT, threshold=HarmBlockThreshold.BLOCK_ONLY_HIGH)
        ]
        """
        self.defaultPrompt = ""
        #self.enableVision = (os.path.realpath(__file__).endswith("vision.py"))
        self.promptStyle = Style.from_dict({
            # User input (default text).
            "": config.terminalCommandEntryColor2,
            # Prompt.
            "indicator": config.terminalPromptIndicatorColor2,
        })

    def setSystemMessage(self):
        # completer
        Plugins.runPlugins()
        completer = FuzzyCompleter(WordCompleter(list(config.predefinedContexts.values()), ignore_case=True))
        # history
        historyFolder = os.path.join(config.localStorage, "history")
        Path(historyFolder).mkdir(parents=True, exist_ok=True)
        system_message_history = os.path.join(historyFolder, "system_message")
        system_message_session = PromptSession(history=FileHistory(system_message_history))
        # prompt
        print2("Change system message below:")
        prompt = SinglePrompt.run(style=self.promptStyle, promptSession=system_message_session, default=config.systemMessage_vertexai, completer=completer)
        if prompt and not prompt == config.exit_entry:
            config.systemMessage_vertexai = prompt
            config.saveConfig()
            print2("System message changed!")

    def run(self, prompt="", once=False):
        if self.defaultPrompt:
            prompt, self.defaultPrompt = self.defaultPrompt, ""
        historyFolder = os.path.join(config.localStorage, "history")
        Path(historyFolder).mkdir(parents=True, exist_ok=True)
        chat_history = os.path.join(historyFolder, "geminipro")
        chat_session = PromptSession(history=FileHistory(chat_history))

        if not self.runnable:
            print(f"{self.name} is not running due to missing configurations!")
            return None
        model = GenerativeModel("gemini-pro")
        # on-going history
        if hasattr(config, "currentMessages") and config.currentMessages:
            history = []
            user = True
            for i in config.currentMessages:
                if i["role"] == "user" if user else "assistant":
                    history.append(Content(role="user" if user else "model", parts=[Part.from_text(i["content"])]))
                    user = not user
            # e.g. history=[
            #    Content(role="user", parts=[Part.from_text("Hello!")]),
            #    Content(role="model", parts=[Part.from_text("Hello! How can I assist you today?")]),
            #]
            if history and history[-1].role == "user":
                history = history[:-1]
            elif not history:
                history = None
        else:
            history = None
        chat = model.start_chat(history=history)
        justStarted = True
        #print2(f"\n{self.name} + Vision loaded!" if self.enableVision else f"\n{self.name} loaded!")
        print2(f"\n{self.name} loaded!")
        print2("```system message")
        print1(config.systemMessage_vertexai)
        print2("```")
        if hasattr(config, "currentMessages"):
            bottom_toolbar = f""" {str(config.hotkey_exit).replace("'", "")} {config.exit_entry}"""
        else:
            bottom_toolbar = f""" {str(config.hotkey_exit).replace("'", "")} {config.exit_entry} {str(config.hotkey_new).replace("'", "")} .new"""
            print("(To start a new chart, enter '.new')")
        print(f"(To exit, enter '{config.exit_entry}')\n")
        while True:
            completer = None if hasattr(config, "currentMessages") else FuzzyCompleter(WordCompleter([".new", ".systemmessage", ".toggleinputaudio", ".toggleoutputaudio", config.exit_entry], ignore_case=True))
            if not prompt:
                prompt = SinglePrompt.run(style=self.promptStyle, promptSession=chat_session, bottom_toolbar=bottom_toolbar, completer=completer)
                if prompt and not prompt in (".new", config.exit_entry) and hasattr(config, "currentMessages"):
                    config.currentMessages.append({"content": prompt, "role": "user"})
            else:
                prompt = SinglePrompt.run(style=self.promptStyle, promptSession=chat_session, bottom_toolbar=bottom_toolbar, default=prompt, accept_default=True, completer=completer)
            if prompt == config.exit_entry:
                break
            elif not hasattr(config, "currentMessages") and prompt.lower() == ".toggleinputaudio":
                toggleinputaudio()
            elif not hasattr(config, "currentMessages") and prompt.lower() == ".toggleoutputaudio":
                toggleoutputaudio()
            elif not hasattr(config, "currentMessages") and prompt.lower() == ".systemmessage":
                self.setSystemMessage()
                clear()
                chat = model.start_chat()
                justStarted = True
                print("New chat started!")
            elif not hasattr(config, "currentMessages") and prompt.lower() == ".new":
                clear()
                chat = model.start_chat()
                justStarted = True
                print("New chat started!")
            elif prompt := prompt.strip():
                streamingWordWrapper = StreamingWordWrapper()

                # declare a function
#                get_vision_func = generative_models.FunctionDeclaration(
#                    name="analyze_images",
#                    description="Describe or analyze images. Remember, do not use this function for non-image related tasks. Even it is an image-related task, use this function ONLY if I provide at least one image file path or image url.",
#                    parameters={
#                        "type": "object",
#                        "properties": {
#                            "query": {
#                                "type": "string",
#                                "description": "Questions or requests that users ask about the given images",
#                            },
#                            "files": {
#                                "type": "string",
#                                "description": """Return a list of image paths or urls, e.g. '["image1.png", "/tmp/image2.png", "https://letmedoit.ai/image.png"]'. Return '[]' if image path is not provided.""",
#                            },
#                        },
#                        "required": ["query", "files"],
#                    },
#                )
#                vision_tool = generative_models.Tool(
#                    function_declarations=[get_vision_func],
#                )

                try:
                    if not hasattr(config, "currentMessages") and config.systemMessage_vertexai and justStarted:
                        prompt = f"# Your role\n\n{config.systemMessage_vertexai}\n\n# My Inquiry\n\n{prompt}"
                        justStarted = False
                    # https://cloud.google.com/vertex-ai/docs/generative-ai/model-reference/gemini
                    # Note: At the time of writing, function call feature with Gemini Pro is very weak, compared with the function call feature offerred by ChatGPT:
                    # 1. Gemini Pro do not accept multiple tools in a single message
                    # 2. Gemini Pro is weak to determine if it is appropriate to use the given tool or not.  When a tool is given, it is called by mistake so often.  In contrast, ChatGPT has the "auto" setting which makes ChatGPT obviously smarter than Gemini Pro.
                    #if "[NO_TOOL]" in prompt or not self.enableVision:
                    #    allow_function_call = False
                    #    prompt = prompt.replace("[NO_TOOL]", "")
                    #else:
                    #    allow_function_call = True
                    completion = chat.send_message(
                        prompt,
                        # Optional:
                        generation_config=self.generation_config,
                        safety_settings=self.safety_settings,
                        #tools=[vision_tool] if allow_function_call else None,
                        stream=True,
                    )

                    # Create a new thread for the streaming task
                    streaming_event = threading.Event()
                    self.streaming_thread = threading.Thread(target=streamingWordWrapper.streamOutputs, args=(streaming_event, completion,))
                    # Start the streaming thread
                    self.streaming_thread.start()

                    # wait while text output is steaming; capture key combo 'ctrl+q' or 'ctrl+z' to stop the streaming
                    streamingWordWrapper.keyToStopStreaming(streaming_event)

                    # when streaming is done or when user press "ctrl+q"
                    self.streaming_thread.join()

                except:
                    self.streaming_thread.join()
                    print2(traceback.format_exc())

            prompt = ""

            if once:
                break

        print2(f"\n{self.name} closed!")
        if hasattr(config, "currentMessages"):
            print2(f"Return back to {config.toolMateAIName} prompt ...")

def main():
    # Create the parser
    parser = argparse.ArgumentParser(description="geminipro cli options")
    # Add arguments
    parser.add_argument("default", nargs="?", default=None, help="default entry")
    parser.add_argument('-o', '--outputtokens', action='store', dest='outputtokens', help="specify maximum output tokens with -o flag; default: 8192")
    parser.add_argument('-t', '--temperature', action='store', dest='temperature', help="specify temperature with -t flag: default: 0.9")
    # Parse arguments
    args = parser.parse_args()
    # Get options
    prompt = args.default.strip() if args.default and args.default.strip() else ""
    if args.outputtokens and args.outputtokens.strip():
        try:
            max_output_tokens = int(args.outputtokens.strip())
        except:
            max_output_tokens = 8192
    else:
        max_output_tokens = 8192
    if args.temperature and args.temperature.strip():
        try:
            temperature = float(args.temperature.strip())
        except:
            temperature = 0.9
    else:
        temperature = 0.9
    GeminiPro(
        temperature=temperature,
        max_output_tokens = max_output_tokens,
    ).run(
        prompt=prompt,
    )

if __name__ == '__main__':
    main()


from toolmate import config

config.predefinedContexts["Think"] = """@chat
# Instructions

Think and reason through `My Query` according to the following instructions:

1. Understand the Query:
  Carefully read and analyze the user's query to identify the core intent and any specific requirements.
  Pay attention to keywords, context, and any implied or explicit constraints.

  a. Identify the main question or task being asked.
  b. Recognize any constraints or limitations mentioned.
  c. Consider the expected output format (e.g., short answer, detailed explanation, list, json, structed layout, etc.).
  d. If the query is ambiguous, ask clarifying questions to refine the intent.

2. Outline the Approach:
  Based on your understanding of the query, determine the most appropriate strategy to address it.
  Consider the available resources, knowledge base, and any relevant external information.

  a. Determine the most suitable methodology or techniques to apply.
  b. Consider whether the query requires:
    i. Factual recall from your knowledge base.
    ii. Logical reasoning and inference.
    iii. Mathematical or computational analysis.
    iv. Creative generation of text or ideas.
  c. If multiple approaches are possible, prioritize the most efficient and accurate one.

3. Present a Plan:
  Break down the approach into a clear sequence of steps.

  a. Clearly outline the steps involved in addressing the query.
  b. Each step should be concise and actionable.
  c. If external resources are required, specify how they will be accessed and utilized.
  d. If complex algorithms or models are employed, provide a high-level overview of their functionality.

4. Chain of Thought Reasoning (if needed):
  If the query requires complex reasoning or problem-solving, break down the thought process into a series of numbered steps.
  Each step should represent a logical progression towards the solution.

  a. Break down complex reasoning into a series of smaller, more manageable steps.
  b. Each step should represent a clear logical inference or deduction.
  c. Number the steps to maintain a clear sequence.
  d. Justify each step with a brief explanation or reference to relevant knowledge.
  e. If assumptions are made, clearly state them and their rationale.

# Output Format

* Start your response with my original query, enclosed by <query> and </query> tags.
* Opening another tag <thinking>, followed by the rest of your response.
* Close your response with the tag </thinking>

# My Query

"""

config.predefinedContexts["Review"] = """@chat
# Instructions

Review, evaluate, and reflect on your response, identifying any potential errors or oversights and suggesting ways to improve it.

1. Context and Original Query:

  a. Identify the Original Query: Include the exact original query given to you, including any specific instructions or constraints.
  b. Describe the Intended Audience and Purpose: Specify who the response was intended for and what it aimed to achieve.

2. Thorough Review of the Response:

  a. Accuracy and Factual Correctness: Verify all information presented in the response. Check for any errors, inconsistencies, or outdated information.
  b. Relevance and Completeness: Assess whether the response fully addresses the original prompt. Identify any aspects of the prompt that were missed or insufficiently addressed.
  c. Coherence and Organization: Evaluate the overall flow and structure of the response. Is it well-organized and easy to follow? Are there any abrupt transitions or disjointed sections?
  d. Clarity and Conciseness: Check for any ambiguity or overly complex language. Ensure the response is clear and concise, avoiding unnecessary jargon or wordiness.
  e. Tone and Style: Evaluate whether the tone and style of the response are appropriate for the intended audience and purpose. Is it formal, informal, persuasive, informative, etc.?

3. Potential Errors and Oversights:

  a. Identify Factual Errors: Point out any specific instances where the LLM response provides incorrect or misleading information.
  b. Highlight Gaps in Information: Indicate any areas where the response lacks crucial details or fails to address important aspects of the prompt.
  c. Address Logical Fallacies: If the LLM response contains any flawed reasoning or logical inconsistencies, clearly identify and explain them.
  d. Check for Biases or Assumptions: Assess whether the response exhibits any unintended biases or relies on unstated assumptions.

4. Suggestions for Improvement:

  a. Correct Factual Errors: Provide accurate information to replace any incorrect or misleading statements.
  b. Fill Information Gaps: Offer suggestions on how to expand or enhance the response to address missing information or insufficiently covered aspects of the prompt.
  c. Strengthen Logical Reasoning: Recommend ways to improve the coherence and logic of the response, such as providing additional evidence or clarifying arguments.
  d. Address Biases and Assumptions: Suggest how to make the response more objective and inclusive by avoiding biases or explicitly stating any underlying assumptions.
  e. Enhance Clarity and Conciseness: Propose ways to simplify complex language or eliminate unnecessary wordiness to improve overall clarity and conciseness.
  f. Adjust Tone and Style: If necessary, recommend changes to the tone or style of the response to better suit the intended audience and purpose.

5. Overall Assessment and Reflection:

  a. Summarize Strengths and Weaknesses: Provide a concise overview of the LLM response's strengths and weaknesses.
  b. Reflect on Potential Causes of Errors: Consider why the LLM may have made certain errors or oversights. Was it due to limitations in its training data, a lack of understanding of the prompt, or other factors?
  c. Suggest Strategies for Future Improvement: Based on the evaluation and reflection, propose ways to enhance the LLM's performance on similar tasks in the future.

"""

config.predefinedContexts["Refine"] = """@chat
# Instruction

Provide a detailed and improved response to the original query, considering the information discussed thus far.

"""

# Context `Reflection` is modified from ideas presented at:
# https://huggingface.co/mattshumer/Reflection-Llama-3.1-70B
# https://generativeai.pub/how-i-built-a-2b-reflection-llm-surprisingly-good-fae76bc2a0ba
config.predefinedContexts["Reflection"] = """@chat
# Instructions

You are a world-class AI system, capable of complex reasoning and reflection.
You are designed to provide detailed, step-by-step responses to `My Query`. 
Your outputs should follow this structure:
1. Begin with a <thinking> section.
2. Inside the thinking section:
  a. Briefly analyze `My Query` and outline your approach.
  b. Present a clear plan of steps to resolve the query.
  c. Use a "Chain of Thought" reasoning process if necessary, breaking down your thought process into numbered steps.
3. Include a <reflection> section for each idea where you:
  a. Review your reasoning.
  b. Check for potential errors or oversights.
  c. Confirm or adjust your conclusion if necessary.
4. Be sure to close all reflection sections.
5. Close the thinking section with </thinking>.
6. Provide your final answer in an <output> section.
  a. Integrate corrections if there are errors in initial thinking.
  b. Integrate additions if there are oversights in initial thinking.
  c. Refine the whole answer seamlessly.
Always use these tags in your responses. Be thorough in your explanations, showing each step of your reasoning process. 
Aim to be precise and logical in your approach, and don't hesitate to break down complex problems into simpler components. 
Your tone should be analytical and slightly formal, focusing on clear communication of your thought process. 
Remember: Both <thinking> and <reflection> MUST be tags and must be closed at their conclusions. 
Make sure all <tags> are on separate lines with no other text. 
Do not include other text on a line containing a tag.

# My Query

"""

config.inputSuggestions += ["@reflection ", "@deep_reflection "]
config.aliases["@reflection "] = "@chat `Reflection` "
config.builtinTools["reflection"] = "Think and reason through a query, review and refine a response"
config.builtinTools["deep_reflection"] = "Think and reason through a query, review and refine a response in detail"


"""
ToolMate AI Plugin - Ask Perplexica

Default Configurations
(assuming Perplexica is installed locally)
perplexica_server = "localhost"
perplexica_frontend_port = 3000
perplexica_backend_port = 3001

You can manually edit config.py to customise these settings.

[TOOL_CALL]
"""

from toolmate import config, isServerAlive, print1, print2, plainTextToUrl, get_local_ip, getGroqApi_key
import requests, json, re, copy

persistentConfigs = (
    ("perplexica_server", "http://localhost"),
    ("perplexica_frontend_port", 3000),
    ("perplexica_backend_port", 3001),
    ("perplexica_local_embedding_model", "xenova-bge-small-en-v1.5"), # local options "xenova-bge-small-en-v1.5", "xenova-gte-small", "xenova-bert-base-multilingual-uncased"
)
config.setConfig(persistentConfigs)
if config.perplexica_server == "localhost":
    config.perplexica_server = "http://localhost"

if not isServerAlive(re.sub("http://|https://", "", config.perplexica_server), config.perplexica_backend_port):
    config.perplexica_server = f"http://{get_local_ip()}" # access to the server running outside a container

if isServerAlive(re.sub("http://|https://", "", config.perplexica_server), config.perplexica_backend_port):

    def ask_perplexica_chatgpt(function_args):
        config.stopSpinning()
        if function_args:
            query = function_args.get("query")
            config.currentMessages[-1] = {"role": "user", "content": query}
        else:
            query = config.currentMessages[-1]["content"]

        history = []
        for i in config.currentMessages[:-1]:
            role = i.get("role", "")
            if role == "assistant":
                history.append(i)
            elif role == "user":
                i["role"] = "human"
                history.append(i)
        #print(history)

        api_url = f"{config.perplexica_server}:{config.perplexica_backend_port}/api/search" 
        headers = {"Content-Type": "application/json"}
        # references:
        # https://github.com/ItzCrazyKns/Perplexica/blob/master/docs/API/SEARCH.md
        # https://github.com/ItzCrazyKns/Perplexica/tree/master/src/lib/providers

        data = {
            "chatModel": {
                "provider": "openai",
                "model": config.chatGPTApiModel,
            },
            "embeddingModel": {
                "provider": "openai",
                "model": "text-embedding-3-large",
            },
            "optimizationMode": "speed",
            "focusMode": "webSearch",
            "query": query,
            "history": history,
        }

        try:
            response = requests.post(api_url, headers=headers, data=json.dumps(data))
            response.raise_for_status()  # Raise an exception for bad status codes

            response_json = response.json()
            #print(response_json) 

            answer = response_json["message"]
            sources = response_json["sources"]

            config.toolTextOutput = f"```answer\n{answer}\n```\n\n```sources"
            print2("```answer")
            print1(answer)
            print2("```\n\n```sources")

            for index, i in enumerate(sources):
                if "metadata" in i:
                    title = i["metadata"].get("title", "")
                    url = i["metadata"].get("url", "")
                    if title and url:
                        source = f"{(index + 1)}. [{title}]({url})"
                        config.toolTextOutput += f"{source}\n"
                        print(source)
            config.toolTextOutput += "```"
            print2("```")

        except requests.exceptions.RequestException as e:
            print(f"An error occurred: {e}")
        except json.JSONDecodeError as e:
            print(f"Failed to decode JSON response: {e}")
        except KeyError as e:
            print(f"Missing key in response: {e}")
        
        return ""

    functionSignature = {
        "examples": [
            "Ask Perplexica",
        ],
        "name": "ask_perplexica_chatgpt",
        "description": "Request Perplexica to conduct research or provide information through internet searches.",
        "parameters": {
            "type": "object",
            "properties": {} if not config.tool_selection_agent else {
                "query": {
                    "type": "string",
                    "description": "The original request in detail, including any supplementary information",
                },
            },
            "required": [] if not config.tool_selection_agent else ["query"],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=ask_perplexica_chatgpt)
    config.inputSuggestions.append("Ask Perplexica: ")

else:

    print1(f"Perplexica Backend `{config.perplexica_server}:{config.perplexica_backend_port}` not found! Plugin `search perplexica not enabled!`")



"""
ToolMate AI Plugin - read aloud

pronunce words

[TOOL_CALL]
"""

from toolmate import config, print1
from toolmate.utils.tts_utils import TTSUtil
import re

def read_aloud(function_args):
    config.stopSpinning()
    if function_args:
        content = function_args.get("words")
        config.currentMessages[-1] = {"role": "user", "content": content}
    else:
        content = config.currentMessages[-1]["content"]
    content = re.sub("^[Rr]ead aloud[:]*", "", content).strip()
    #words = function_args.get("words") # required
    #language = function_args.get("language") # required
    print1("Loading speech feature ...")
    #TTSUtil.play(words, language)
    TTSUtil.play(content)
    return ""

functionSignature = {
    "examples": [
        "read aloud",
    ],
    "name": "read_aloud",
    "description": "Pronounce words or sentences with text-to-speech utility",
    "parameters": {
        "type": "object",
        "properties": {} if not config.tool_selection_agent else {
            "words": {
                "type": "string",
                "description": "Words to be pronounced",
            },
        },
        "required": [] if not config.tool_selection_agent else ["words"],
    },
}

config.addFunctionCall(signature=functionSignature, method=read_aloud)
config.inputSuggestions.append("Read aloud: ")



"""
ToolMate AI Plugin - install python package

install python package into the environment that runs LetMeDoIt AI

[TOOL_CALL]
"""

from toolmate import config
from toolmate import installPipPackage

# Function method
def install_package(function_args):
    config.stopSpinning()
    if function_args:
        package = function_args.get("package") # required
        #config.currentMessages[-1] = {"role": "user", "content": package}
    else:
        package = config.currentMessages[-1]["content"]
    if package:
        install = installPipPackage(f"--upgrade {package}")
        return "Installed!" if install else f"Failed to install '{package}'!"
    return ""

# Function Signature
functionSignature = {
    "examples": [
        "install python package",
    ],
    "name": "install_package",
    "description": f'''Install a python package''',
    "parameters": {
        "type": "object",
        "properties": {} if not config.tool_selection_agent else {
            "package": {
                "type": "string",
                "description": "Package name",
            },
        },
        "required": [] if not config.tool_selection_agent else ["package"],
    },
}

# Integrate the signature and method into LetMeDoIt AI
config.addFunctionCall(signature=functionSignature, method=install_package)


"""
ToolMate AI Plugin - send sms

send sms on Android

[TOOL_CALL]
"""

if config.isTermux:

    from toolmate import config
    import subprocess

    def send_sms(function_args):
        phone_number = function_args.get("phone_number", "") # required
        message = function_args.get("message", "").replace('"', '\\"') # required
        cli = f'''termux-sms-send -n {phone_number} "{message}"'''
        subprocess.Popen(cli, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return ""

    functionSignature = {
        "examples": [],
        "name": "send_sms",
        "description": f'''Send sms''',
        "parameters": {
            "type": "object",
            "properties": {
                "phone_number": {
                    "type": "string",
                    "description": "The phone number of the selected person",
                },
                "message": {
                    "type": "string",
                    "description": "Generate SMS message",
                },
            },
            "required": ["phone_number", "message"],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=send_sms)

    config.aliases["@read_sms "] = "@command termux-sms-list "
    config.builtinTools["read_sms"] = "read sms"
    config.inputSuggestions.append("@read_sms ")


"""
ToolMate AI Plugin - send whatsapp messages

send whatsapp messages on Android

[TOOL_CALL]
"""

if config.isTermux:

    from toolmate import config, stopSpinning
    import subprocess

    def send_whatsapp(function_args):
        message = function_args.get("message").replace('"', '\\"') # required
        config.stopSpinning()
        # e.g. am start -a android.intent.action.VIEW -d "https://api.whatsapp.com/send?phone=+441234567&text=Hello"
        cli = f'''am start -a android.intent.action.VIEW -d "https://api.whatsapp.com/send?text={message}"'''
        subprocess.Popen(cli, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return ""

    functionSignature = {
        "examples": [
            "send WhatsApp",
        ],
        "name": "send_whatsapp",
        "description": f'''Send WhatsApp messages''',
        "parameters": {
            "type": "object",
            "properties": {
                "message": {
                    "type": "string",
                    "description": "The message that is to be sent to the recipient",
                },
            },
            "required": ["message"],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=send_whatsapp)


"""
ToolMate AI Plugin - ask gemini pro

Ask Google Gemini Pro for information

[TOOL_CALL]
"""

if os.environ["GOOGLE_APPLICATION_CREDENTIALS"] and "Vertex AI" in config.enabledGoogleAPIs and not config.isLite:

    import vertexai
    from vertexai.generative_models._generative_models import (
        HarmCategory,
        HarmBlockThreshold,
    )
    from toolmate import config
    from toolmate.utils.call_gemini import CallVertexAI

    def ask_gemini(function_args):
        config.stopSpinning()
        if function_args:
            query = function_args.get("query") # required
            config.currentMessages[-1] = {"role": "user", "content": query}
        else:
            query = config.currentMessages[-1]["content"]
        completion = CallVertexAI.regularCall(config.currentMessages)
        config.toolmate.streamCompletion(completion, openai=False)
        return ""

    functionSignature = {
        "examples": [
            "Ask Gemini",
        ],
        "name": "ask_gemini",
        "description": "Ask Gemini to chat or provide information",
        "parameters": {
            "type": "object",
            "properties": {} if not config.tool_selection_agent else {
                "query": {
                    "type": "string",
                    "description": "The original request in detail, including any supplementary information",
                },
            },
            "required": [] if not config.tool_selection_agent else ["query"],
        },
    }

    # initiation
    vertexai.init()

    # Note: BLOCK_NONE is not allowed
    config.gemini_safety_settings={
        HarmCategory.HARM_CATEGORY_UNSPECIFIED: HarmBlockThreshold.BLOCK_ONLY_HIGH,
        HarmCategory.HARM_CATEGORY_DANGEROUS_CONTENT: HarmBlockThreshold.BLOCK_ONLY_HIGH,
        HarmCategory.HARM_CATEGORY_HATE_SPEECH: HarmBlockThreshold.BLOCK_ONLY_HIGH,
        HarmCategory.HARM_CATEGORY_HARASSMENT: HarmBlockThreshold.BLOCK_ONLY_HIGH,
        HarmCategory.HARM_CATEGORY_SEXUALLY_EXPLICIT: HarmBlockThreshold.BLOCK_ONLY_HIGH,
    }
    config.addFunctionCall(signature=functionSignature, method=ask_gemini)
    config.inputSuggestions.append("Ask Gemini: ")
else:
    print("Plugin `ask gemini` not enabled! Read setup at https://github.com/eliranwong/toolmate/blob/main/package/toolmate/docs/Google%20Cloud%20Service%20Credential%20Setup.md")


import textwrap
from toolmate import config
from toolmate import print2, print3, voiceTyping

from prompt_toolkit import prompt
from prompt_toolkit.filters import Condition
from prompt_toolkit.key_binding import KeyBindings, merge_key_bindings, ConditionalKeyBindings
from prompt_toolkit.clipboard.pyperclip import PyperclipClipboard
from prompt_toolkit.application import run_in_terminal


class SinglePrompt:

    @staticmethod
    def run(inputIndicator="", validator=None, default="", accept_default=False, completer=None, promptSession=None, style=None, is_password=False, bottom_toolbar=None, **kwargs):
        this_key_bindings = KeyBindings()
        @this_key_bindings.add(*config.hotkey_exit)
        def _(event):
            buffer = event.app.current_buffer
            buffer.text = config.exit_entry
            buffer.validate_and_handle()
        @this_key_bindings.add(*config.hotkey_cancel)
        def _(event):
            buffer = event.app.current_buffer
            buffer.reset()
        @this_key_bindings.add(*config.hotkey_insert_newline)
        def _(event):
            buffer = event.app.current_buffer
            buffer.newline()
        @this_key_bindings.add(*config.hotkey_toggle_word_wrap)
        def _(_):
            config.wrapWords = not config.wrapWords
            config.saveConfig()
            run_in_terminal(lambda: print3(f"Word Wrap: '{'enabled' if config.wrapWords else 'disabled'}'!"))
        @this_key_bindings.add(*config.hotkey_toggle_response_audio)
        def _(_):
            if config.tts:
                config.ttsOutput = not config.ttsOutput
                config.saveConfig()
                run_in_terminal(lambda: print3(f"Response Audio: '{'enabled' if config.ttsOutput else 'disabled'}'!"))
        @this_key_bindings.add(*config.hotkey_voice_entry)
        def _(event):
            if config.pyaudioInstalled:
                buffer = event.app.current_buffer
                buffer.text = f"{buffer.text}{' ' if buffer.text else ''}{voiceTyping()}"
                if config.voiceTypingAutoComplete:
                    buffer.validate_and_handle()
                else:
                    buffer.cursor_position = buffer.cursor_position + buffer.document.get_end_of_line_position()
            else:
                run_in_terminal(lambda: print2("Install PyAudio first to enable voice recognition!"))

        if hasattr(config, "currentMessages"):
            from toolmate.utils.prompt_shared_key_bindings import prompt_shared_key_bindings
            from toolmate.utils.prompt_multiline_shared_key_bindings import prompt_multiline_shared_key_bindings
            @this_key_bindings.add(*config.hotkey_launch_pager_view)
            def _(_):
                config.launchPager()
            # additional key binding
            conditional_prompt_multiline_shared_key_bindings = ConditionalKeyBindings(
                key_bindings=prompt_multiline_shared_key_bindings,
                filter=Condition(lambda: config.multilineInput),
            )
            this_key_bindings = merge_key_bindings([
                this_key_bindings,
                prompt_shared_key_bindings,
                conditional_prompt_multiline_shared_key_bindings,
            ])
        else:
            @this_key_bindings.add(*config.hotkey_new)
            def _(event):
                buffer = event.app.current_buffer
                config.defaultEntry = buffer.text
                buffer.text = ".new"
                buffer.validate_and_handle()

        config.selectAll = False
        inputPrompt = promptSession.prompt if promptSession is not None else prompt
        if not hasattr(config, "clipboard"):
            config.clipboard = PyperclipClipboard()
        if not inputIndicator:
            inputIndicator = [
                ("class:indicator", ">>> "),
            ]
        userInput = inputPrompt(
            inputIndicator,
            key_bindings=this_key_bindings,
            bottom_toolbar=bottom_toolbar if bottom_toolbar is not None else f""" {str(config.hotkey_exit).replace("'", "")} {config.exit_entry}""",
            #enable_system_prompt=True,
            swap_light_and_dark_colors=Condition(lambda: not config.terminalResourceLinkColor.startswith("ansibright")),
            style=style,
            validator=validator,
            multiline=Condition(lambda: hasattr(config, "currentMessages") and config.multilineInput),
            default=default,
            accept_default=accept_default,
            completer=completer,
            is_password=is_password,
            mouse_support=Condition(lambda: config.mouseSupport),
            clipboard=config.clipboard,
            **kwargs,
        )
        userInput = textwrap.dedent(userInput) # dedent to work with code block
        return userInput if hasattr(config, "addPathAt") and config.addPathAt else userInput.strip()



from toolmate import config, isCommandInstalled, print2
import os, shutil, subprocess

# add config item `fabricPath`.  Users can customise fabric path by editing its value in `config.py`.
#whichFabric = shutil.which("fabric")
persistentConfigs = (
    #("fabricPath", whichFabric if whichFabric else "fabric"),
    ("fabricPath", "fabric"), # more flexible when two packages shared the same user directory
)
config.setConfig(persistentConfigs)

if isCommandInstalled(config.fabricPath) or os.path.isfile(config.fabricPath):
    # add aliases
    config.aliases["@fabric_pattern "] = f"@command {config.fabricPath} -p "
    config.aliases["@append_fabric_pattern "] = f"@append_command {config.fabricPath} -p "
    config.aliases["@fabric "] = f"@command {config.fabricPath} "
    config.aliases["@append_fabric "] = f"@append_command {config.fabricPath} "
    # add input suggestions
    patternsOutput = subprocess.run("fabric -l", shell=True, capture_output=True, text=True).stdout
    patterns = patternsOutput.split("\n\t")[1:] if "\n\tai" in patternsOutput else patternsOutput.split("\n")
    patterns = {f"{i} ": None for i in patterns}
    config.inputSuggestions += [{"@fabric_pattern": patterns}, {"@append_fabric_pattern": patterns}]
    patterns = {"-p": patterns}
    config.inputSuggestions += [{"@fabric": patterns}, {"@append_fabric": patterns}]

    config.builtinTools["fabric"] = "Execute a fabric command"
    config.builtinTools["append_fabric"] = "Execute a fabric command with the previous text output appended to it"
else:
    print2("`fabric` not found! Read https://github.com/danielmiessler/fabric for installation!")



from toolmate import config

# @show_location @show_connection @start_recording @stop_recording
if config.isTermux:
    config.aliases["@show_location "] = "@command termux-location "
    config.aliases["@show_connection "] = "@command termux-wifi-connectioninfo "
    config.aliases["@start_recording "] = "@command termux-microphone-record -d "
    config.aliases["@stop_recording "] = "@command termux-microphone-record -q "

    config.builtinTools["show_location"] = "show location"
    config.builtinTools["show_connection"] = "show connection"
    config.builtinTools["start_recording"] = "start recording"
    config.builtinTools["stop_recording"] = "stop recording"

    config.inputSuggestions += ["@show_location ", "@show_connection ", "@start_recording ", "@stop_recording "]


"""
ToolMate AI Plugin - create images

generate images with model "Google Imagen 3"

[TOOL_CALL]
"""

if not config.isLite:

  from toolmate import config, print3, getCurrentDateTime
  import os, subprocess, shutil
  from pathlib import Path

  import vertexai
  from vertexai.preview.vision_models import ImageGenerationModel


  def create_image_imagen3(function_args):
      def openImageFile(imageFile):
          if config.terminalEnableTermuxAPI:
              getCliOutput(f"termux-share {imageFile}")
          elif shutil.which(config.open):
              cli = f"{config.open} {imageFile}"
              #os.system(cli)
              subprocess.Popen(cli, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

      prompt = function_args.get("prompt") # required

      if os.environ["GOOGLE_APPLICATION_CREDENTIALS"] and "Vertex AI" in config.enabledGoogleAPIs:
          # initiation
          vertexai.init()
      else:
          print("Vertex AI is not enabled!")
          print("Read https://github.com/eliranwong/toolmate/blob/main/package/toolmate/docs/Google%20Cloud%20Service%20Credential%20Setup.md for setting up Google API.")
          return "[INVALID]"
      
      generation_model = ImageGenerationModel.from_pretrained("imagen-3.0-generate-001")

      # reference: https://cloud.google.com/blog/products/ai-machine-learning/a-developers-guide-to-imagen-3-on-vertex-ai?e=0?utm_source%3Dlinkedin
      """Generates images from text prompt.

      Args:
          prompt: Text prompt for the image.
          negative_prompt: A description of what you want to omit in the generated
            images.
          number_of_images: Number of images to generate. Range: 1..8.
          width: Width of the image. One of the sizes must be 256 or 1024.
          height: Height of the image. One of the sizes must be 256 or 1024.
          aspect_ratio: Aspect ratio for the image. Supported values are:
              * 1:1 - Square image
              * 9:16 - Portait image
              * 16:9 - Landscape image
              * 4:3 - Landscape, desktop ratio.
              * 3:4 - Portrait, desktop ratio
          guidance_scale: Controls the strength of the prompt. Suggested values
            are - * 0-9 (low strength) * 10-20 (medium strength) * 21+ (high
            strength)
          seed: Image generation random seed.
          base_image: Base image to use for the image generation.
          mask: Mask for the base image.
          edit_mode: Describes the editing mode for the request. Supported values
            are - * inpainting-insert: fills the mask area based on the text
            prompt (requires mask and text) * inpainting-remove: removes the
            object(s) in the mask area. (requires mask)
              * outpainting: extend the image based on the mask area. (Requires
                mask) * product-image: Changes the background for the predominant
                product or subject in the image
          mask_mode: Solicits generation of the mask (v/s providing mask as an
            input). Supported values are:
              * background: Automatically generates a mask for all regions except
                the primary subject(s) of the image
              * foreground: Automatically generates a mask for the primary
                subjects(s) of the image.
              * semantic: Segment one or more of the segmentation classes using
                class ID
          segmentation_classes: List of class IDs for segmentation. Max of 5 IDs
          mask_dilation: Defines the dilation percentage of the mask provided.
            Float between 0 and 1. Defaults to 0.03
          product_position: Defines whether the product should stay fixed or be
            repositioned. Supported Values:
              * fixed: Fixed position
              * reposition: Can be moved (default)
          output_mime_type: Which image format should the output be saved as.
            Supported values: * image/png: Save as a PNG image * image/jpeg: Save
            as a JPEG image
          compression_quality: Level of compression if the output mime type is
            selected to be image/jpeg. Float between 0 to 100
          language: Language of the text prompt for the image. Default: None.
            Supported values are `"en"` for English, `"hi"` for Hindi, `"ja"` for
            Japanese, `"ko"` for Korean, and `"auto"` for automatic language
            detection.
          output_gcs_uri: Google Cloud Storage uri to store the generated images.
          add_watermark: Add a watermark to the generated image
          safety_filter_level: Adds a filter level to Safety filtering. Supported
            values are: * "block_most" : Strongest filtering level, most strict
            blocking * "block_some" : Block some problematic prompts and responses
            * "block_few" : Block fewer problematic prompts and responses *
            "block_fewest" : Block very few problematic prompts and responses
          person_generation: Allow generation of people by the model Supported
            values are: * "dont_allow" : Block generation of people *
            "allow_adult" : Generate adults, but not children * "allow_all" :
            Generate adults and children

      Returns:
          An `ImageGenerationResponse` object.
      """

      image = generation_model.generate_images(
          prompt=prompt,
          number_of_images=1,
          aspect_ratio="1:1",
          safety_filter_level="block_few", # Literal["block_most", "block_some", "block_few", "block_fewest"]
          person_generation="allow_adult", # Literal["dont_allow", "allow_adult", "allow_all"]
      )
      #print(dir(image[0])) # ['__annotations__', '__class__', '__delattr__', '__dict__', '__dir__', '__doc__', '__eq__', '__format__', '__ge__', '__getattribute__', '__gt__', '__hash__', '__init__', '__init_subclass__', '__le__', '__lt__', '__module__', '__ne__', '__new__', '__reduce__', '__reduce_ex__', '__repr__', '__setattr__', '__sizeof__', '__str__', '__subclasshook__', '__weakref__', '_as_base64_string', '_blob', '_gcs_uri', '_generation_parameters', '_image_bytes', '_loaded_bytes', '_loaded_image', '_mime_type', '_pil_image', '_size', 'generation_parameters', 'load_from_file', 'save', 'show']
      
      # image file path
      folder = os.path.join(config.localStorage, "images")
      Path(folder).mkdir(parents=True, exist_ok=True)
      imageFile = os.path.join(folder, f"{getCurrentDateTime()}.png")
      # save image
      image[0].save(imageFile)
      message = f"Image saved: {imageFile}"
      config.toolTextOutput = message
      print3(message)
      # open image
      openImageFile(imageFile)
      return ""

  functionSignature = {
      "examples": [
          "generate image",
          "create image",
      ],
      "name": "create_image_imagen3",
      "description": "Create an image with Imagen 3",
      "parameters": {
          "type": "object",
          "properties": {
              "prompt": {
                  "type": "string",
                  "description": "Description of the image in as much detail as possible",
              },
          },
          "required": ["prompt"],
      },
  }

  config.addFunctionCall(signature=functionSignature, method=create_image_imagen3)


"""
ToolMate AI Plugin - search SearXNG

Default Configurations
(assuming Perplexica is installed locally)
searx_server = "localhost"
searx_port = 4000

You can manually edit config.py to customise these settings.

[TOOL_CALL]
"""

from toolmate import config, isServerAlive, print1, print2, plainTextToUrl, convert_html_to_markdown
import re, requests

persistentConfigs = (
    ("searx_server", "http://localhost"),
    ("searx_port", 4000),
)
config.setConfig(persistentConfigs)
if config.searx_server == "localhost":
    config.searx_server = "http://localhost"

if not isServerAlive(re.sub("http://|https://", "", config.searx_server), config.searx_port):
    config.searx_server = f"http://{get_local_ip()}" # access to the server running outside a container

if isServerAlive(re.sub("http://|https://", "", config.searx_server), config.searx_port):

    temporaryConfigs = (
        # tabs: https://docs.searxng.org/user/configured_engines.html
        ("searx_tabs", [
            "!general ",
            "!translate ",
            "!web ",
            "!wikimedia ",
            "!images ",
            "!web ",
            "!videos ",
            "!web ",
            "!news ",
            "!web ",
            "!wikimedia ",
            "!map ",
            "!music ",
            "!lyrics ",
            "!radio ",
            "!it ",
            "!packages ",
            "!q&a ",
            "!repos ",
            "!software_wikis ",
            "!science ",
            "!scientific_publications ",
            "!wikimedia ",
            "!files ",
            "!apps ",
            "!social_media ",
        ]),
        ("searx_categories", []),
    )
    config.setConfig(temporaryConfigs, temporary=True)

    def refineSearchResults(content, removeLink=True):
        content = convert_html_to_markdown(content)
        content = re.sub(r"\nNext page\n[\d\D]*$", "", content) # trim the footer
        searchResults = content.split("\n#")[1:] # trim the header
        refinedSearchResults = []
        for index, result in enumerate(searchResults):
            paragraphs = result.split("\n\n")
            textOnlyParagraphs = []
            for paragraph in paragraphs:
                if not re.search(r"\n\[[^\[\]]+\]\([^\)\)]+\)\n", f"\n{paragraph.strip()}\n"):
                    textOnlyParagraphs.append(paragraph.strip())
            if textOnlyParagraphs:
                refinedSearchResults.append("\n\n".join(textOnlyParagraphs))
            elif strippedResult := result.strip():
                refinedSearchResults.append(strippedResult)
            if len(refinedSearchResults) >= config.maximumInternetSearchResults:
                break
        return "#" + "\n\n#".join(refinedSearchResults)

    def search_searxng(function_args):
        if function_args:
            query = function_args.get("query")
            config.currentMessages[-1] = {"role": "user", "content": query}
        else:
            query = config.currentMessages[-1]["content"]
        if config.searx_categories:
            config.searx_categories = [i[1:] for i in config.searx_categories]
            categories = ",".join(config.searx_categories)
        #from langchain_community.utilities import SearxSearchWrapper
        #context = SearxSearchWrapper(searx_host=f"{config.searx_server}:{config.searx_port}").run(query, categories=config.searx_categories if config.searx_categories else None)
        fullUrl = f"{config.searx_server}:{config.searx_port}/search?q={plainTextToUrl(query)}&categories={categories}" if config.searx_categories else f"{config.searx_server}:{config.searx_port}/search?q={plainTextToUrl(query)}"
        context = requests.get(fullUrl, timeout=60).text
        context = refineSearchResults(context) if context else ""
        config.searx_categories = []
        config.stopSpinning()
        print2("```url")
        try:
            print1(fullUrl)
        except:
            print(fullUrl)
        print2("```")
        return context

    functionSignature = {
        "examples": [
            "Search SearXNG",
        ],
        "name": "search_searxng",
        "description": "Perform online searches to obtain the latest and most up-to-date, real-time information",
        "parameters": {
            "type": "object",
            "properties": {} if not config.tool_selection_agent else {
                "query": {
                    "type": "string",
                    "description": "The original request in detail, including any supplementary information",
                },
            },
            "required": [] if not config.tool_selection_agent else ["query"],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=search_searxng)
    config.aliases["@ask_internet "] = "@search_searxng "
    for i in config.searx_tabs:
        tool = i[1:-1]
        config.aliases[f'@{"qna" if tool=="q&a" else tool} '] = f"@search_searxng {i}"
        config.inputSuggestions.append(f'@{"qna" if tool=="q&a" else tool} ')
        config.builtinTools["qna" if tool=="q&a" else tool] = f"""Search the '{"questions_and_answers" if tool=="q&a" else tool}' category for online information."""
    tabsDict = {i: None for i in config.searx_tabs}
    config.inputSuggestions += ["Search SearxNG: ", {"@search_searxng": tabsDict}, {"@ask_internet": tabsDict}]

else:

    print1(f"Searx Host `{config.searx_server}:{config.searx_port}` not found! Plugin `search searxng not enabled!`")


import vertexai, os, traceback, argparse
from vertexai.language_models import CodeChatModel, ChatMessage
from toolmate import config, getPygmentsStyle, startSpinning, stopSpinning
from toolmate import print1, print2, print3, toggleinputaudio, toggleoutputaudio
from toolmate.utils.single_prompt import SinglePrompt
from toolmate.utils.tool_plugins import Plugins
import pygments
from pygments.lexers.markup import MarkdownLexer
from prompt_toolkit.completion import WordCompleter, FuzzyCompleter
from prompt_toolkit.formatted_text import PygmentsTokens
from prompt_toolkit import print_formatted_text
from prompt_toolkit.styles import Style
from prompt_toolkit import PromptSession
from prompt_toolkit.history import FileHistory
from prompt_toolkit.shortcuts import clear
from pathlib import Path


# Install google-cloud-aiplatform FIRST!
#!pip install --upgrade google-cloud-aiplatform


class Codey:

    def __init__(self, name="Codey"):
        # authentication
        if os.environ["GOOGLE_APPLICATION_CREDENTIALS"] and "Vertex AI" in config.enabledGoogleAPIs:
            self.runnable = True
        else:
            print("Vertex AI is not enabled!")
            print("Read https://github.com/eliranwong/toolmate/blob/main/package/toolmate/docs/Google%20Cloud%20Service%20Credential%20Setup.md for setting up Google API.")
            self.runnable = False
        # initiation
        vertexai.init()
        self.name = name
        self.promptStyle = Style.from_dict({
            # User input (default text).
            "": config.terminalCommandEntryColor2,
            # Prompt.
            "indicator": config.terminalPromptIndicatorColor2,
        })

    def setSystemMessage(self):
        # completer
        Plugins.runPlugins()
        completer = FuzzyCompleter(WordCompleter(list(config.predefinedContexts.values()), ignore_case=True))
        # history
        historyFolder = os.path.join(config.localStorage, "history")
        Path(historyFolder).mkdir(parents=True, exist_ok=True)
        system_message_history = os.path.join(historyFolder, "system_message")
        system_message_session = PromptSession(history=FileHistory(system_message_history))
        # prompt
        print2("Change system message below:")
        prompt = SinglePrompt.run(style=self.promptStyle, promptSession=system_message_session, default=config.systemMessage_codey, completer=completer)
        if prompt and not prompt == config.exit_entry:
            config.systemMessage_codey = prompt
            config.saveConfig()
            print2("System message changed!")

    def run(self, prompt="", model="codechat-bison-32k", temperature=0.2, max_output_tokens=2048, once=False):
        historyFolder = os.path.join(config.localStorage, "history")
        Path(historyFolder).mkdir(parents=True, exist_ok=True)
        chat_history = os.path.join(historyFolder, self.name.replace(" ", "_"))
        chat_session = PromptSession(history=FileHistory(chat_history))

        if not self.runnable:
            print(f"{self.name} is not running due to missing configurations!")
            return None
        model = CodeChatModel.from_pretrained(model)
        # https://cloud.google.com/vertex-ai/docs/generative-ai/model-reference/code-chat
        parameters = {
            "temperature": temperature,  # Temperature controls the degree of randomness in token selection; 0.0–1.0; Default: 0.2
            "max_output_tokens": max_output_tokens,  # Token limit determines the maximum amount of text output; 1–2048; Default: 1024
        }
        # chat history
        if hasattr(config, "currentMessages") and config.currentMessages:
            history = []
            user = True
            for i in config.currentMessages:
                if i["role"] == "user" if user else "assistant":
                    history.append(ChatMessage(content=i["content"], author="user" if user else "model"))
                    user = not user
            if history and history[-1].author == "user":
                history = history[:-1]
            elif not history:
                history = None
        else:
            history = None
        # start chat
        chat = model.start_chat(
            context=config.systemMessage_codey,
            message_history=history,
        )
        print2(f"\n{self.name} loaded!")
        print2("```system message")
        print1(config.systemMessage_codey)
        print2("```")
        if hasattr(config, "currentMessages"):
            bottom_toolbar = f""" {str(config.hotkey_exit).replace("'", "")} {config.exit_entry}"""
        else:
            bottom_toolbar = f""" {str(config.hotkey_exit).replace("'", "")} {config.exit_entry} {str(config.hotkey_new).replace("'", "")} .new"""
            print("(To start a new chart, enter '.new')")
        print(f"(To exit, enter '{config.exit_entry}')\n")
        while True:
            completer = None if hasattr(config, "currentMessages") else FuzzyCompleter(WordCompleter([".new", ".systemmessage", ".toggleinputaudio", ".toggleoutputaudio", config.exit_entry], ignore_case=True))
            if not prompt:
                prompt = SinglePrompt.run(style=self.promptStyle, promptSession=chat_session, bottom_toolbar=bottom_toolbar, completer=completer)
                if prompt and not prompt in (".new", config.exit_entry) and hasattr(config, "currentMessages"):
                    config.currentMessages.append({"content": prompt, "role": "user"})
            else:
                prompt = SinglePrompt.run(style=self.promptStyle, promptSession=chat_session, bottom_toolbar=bottom_toolbar, default=prompt, accept_default=True, completer=completer)
            if prompt == config.exit_entry:
                break
            elif not hasattr(config, "currentMessages") and prompt.lower() == ".toggleinputaudio":
                toggleinputaudio()
            elif not hasattr(config, "currentMessages") and prompt.lower() == ".toggleoutputaudio":
                toggleoutputaudio()
            elif not hasattr(config, "currentMessages") and prompt.lower() == ".systemmessage":
                self.setSystemMessage()
                clear()
                chat = model.start_chat()
                print("New chat started!")
            elif not hasattr(config, "currentMessages") and prompt.lower() == ".new":
                clear()
                chat = model.start_chat()
                print("New chat started!")
            elif prompt := prompt.strip():
                try:
                    startSpinning()
                    response = chat.send_message(
                        prompt, **parameters
                    )
                    stopSpinning()
                    chatResponse = response.text.strip()
                    # color response with markdown style
                    tokens = list(pygments.lex(chatResponse, lexer=MarkdownLexer()))
                    print_formatted_text(PygmentsTokens(tokens), style=getPygmentsStyle())
                    # integrate messages into LetMeDoIt messages
                    if hasattr(config, "currentMessages") and chatResponse:
                        config.currentMessages.append({"role": "assistant", "content": chatResponse})
                except:
                    stopSpinning()
                    print2(traceback.format_exc())

            prompt = ""

            if once:
                break

        print2(f"\n{self.name} closed!")
        if hasattr(config, "currentMessages"):
            print2(f"Return back to {config.toolMateAIName} prompt ...")

def main():
    # Create the parser
    parser = argparse.ArgumentParser(description="codey cli options")
    # Add arguments
    parser.add_argument("default", nargs="?", default=None, help="default entry")
    parser.add_argument('-m', '--model', action='store', dest='model', help="specify language model with -m flag; default: codechat-bison-32k")
    parser.add_argument('-o', '--outputtokens', action='store', dest='outputtokens', help="specify maximum output tokens with -o flag; default: 2048")
    parser.add_argument('-t', '--temperature', action='store', dest='temperature', help="specify temperature with -t flag; default: 0.2")
    # Parse arguments
    args = parser.parse_args()
    # Get options
    prompt = args.default.strip() if args.default and args.default.strip() else ""
    model = args.model.strip() if args.model and args.model.strip() else "codechat-bison-32k"
    if args.outputtokens and args.outputtokens.strip():
        try:
            max_output_tokens = int(args.outputtokens.strip())
        except:
            max_output_tokens = 2048
    else:
        max_output_tokens = 2048
    if args.temperature and args.temperature.strip():
        try:
            temperature = float(args.temperature.strip())
        except:
            temperature = 0.2
    else:
        temperature = 0.2

    Codey().run(
        prompt=prompt,
        model=model,
        temperature=temperature,
        max_output_tokens = max_output_tokens,
    )

if __name__ == '__main__':
    main()


"""
ToolMate AI Plugin - analyze images

analyze images

Platform: llamacpp, ollama
Model: llava <- customizable
To customise:
Change in config.py:
llamacppVisionModel_model_path
llamacppVisionModel_clip_model_path
ollamaVisionModel

Platform: gemini
Model: Gemini Pro Vision

Platform: chaptgpt, letmedoit
Model "gpt-4o"
Reference: https://platform.openai.com/docs/guides/vision

[TOOL_CALL]
"""

if not config.isLite:

    from toolmate import config, print1, print2, is_valid_image_file, is_valid_image_url, startLlamacppVisionServer, stopLlamacppVisionServer, is_valid_url, runToolMateCommand, getLlamacppServerClient
    from toolmate.utils.call_chatgpt import check_openai_errors
    import os
    from openai import OpenAI
    from toolmate.geminiprovision import GeminiProVision
    from toolmate.utils.call_ollama import CallOllama

    @check_openai_errors
    def analyze_images_gemini(function_args):
        from toolmate import config

        answer = GeminiProVision(temperature=config.llmTemperature).analyze_images_gemini(function_args)
        if answer:
            config.toolTextOutput = answer
            return ""
        else:
            return "[INVALID]"

    functionSignature = {
        "examples": [
            "describe image",
            "compare images",
            "analyze image",
        ],
        "name": "analyze_images_gemini",
        "description": "Describe or compare images with Gemini",
        "parameters": {
            "type": "object",
            "properties": {
                "query": {
                    "type": "string",
                    "description": "Questions or requests that users ask about the given images",
                },
                "image_filepath": {
                    "type": "string",
                    "description": """Return a list of image paths or urls, e.g. '["image1.png", "/tmp/image2.png", "https://letmedoit.ai/image.png"]'. Return '[]' if image path is not provided.""",
                },
            },
            "required": ["query", "image_filepath"],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=analyze_images_gemini)



import requests, argparse, json, sys, os
from toolmate import config

# Create the parser
parser = argparse.ArgumentParser(description="ToolMate AI API client cli options")
# Add arguments
parser.add_argument("default", nargs="?", default=None, help="instruction sent to ToolMate API server; use `.` to display the last response only")
parser.add_argument('-a', '--all', action='store', dest='all', help="display the whole conversation; either 'plain' or 'json'; submit `.` as instruction to display conversation only, e.g. `-a plain .`")
parser.add_argument('-c', '--chat', action='store', dest='chat', help="set 'true' to chat as an on-going conversation")
parser.add_argument('-f', '--chatfile', action='store', dest='chatfile', help="specify the file path of a saved conversation")
parser.add_argument('-k', '--key', action='store', dest='key', help="API key")
parser.add_argument('-p', '--port', action='store', dest='port', help="server port")
parser.add_argument('-s', '--server', action='store', dest='server', help="server address; 'http://localhost' by default")
parser.add_argument('-t', '--tools', action='store', dest='tools', help="search enabled tools; use '@' to display all; use regex pattern to filter")
# Parse arguments
args = parser.parse_args()

def chat():
    main(True if not (args.chat is not None and args.chat.lower() == "false") else False)

def main(chat: bool = False):
    cliDefault = args.default.strip() if args.default is not None and args.default.strip() else ""
    stdin_text = sys.stdin.read() if not sys.stdin.isatty() else ""

    if args.tools is not None and args.tools.strip(): # -t given; search tools; ignore all other arguments
        query = args.tools.strip().lower()
        endpoint = f"{args.server if args.server else config.toolmate_api_client_host}:{args.port if args.port else config.toolmate_api_client_port}/api/tools"

        url = f"""{endpoint}?query={query}"""
        headers = {
            "Content-Type": "application/json",
            "X-API-Key": args.key if args.key else config.toolmate_api_client_key,
        }
        response = requests.post(url, headers=headers)
        print(json.loads(response.json())["tools"])

    else: # default given; "." for display current conversation only
        query = cliDefault + stdin_text
        if not query:
            query = "."
        endpoint = f"{args.server if args.server else config.toolmate_api_client_host}:{args.port if args.port else config.toolmate_api_client_port}/api/toolmate"
        if args.chat is not None and args.chat.lower() == "true":
            chat = True
        chatfile = f"&chatfile={args.chatfile}" if args.chatfile is not None and os.path.isfile(args.chatfile) else ""

        url = f"""{endpoint}?cwd={os.getcwd()}&chat={chat}{chatfile}&query={query}"""
        headers = {
            "Content-Type": "application/json",
            "X-API-Key": args.key if args.key else config.toolmate_api_client_key,
        }
        response = requests.post(url, headers=headers)

        #response = requests.post(endpoint, json=prompt)
        if args.all and args.all.lower() in ("plain", "json"):
            if args.all.lower() == "plain":
                for i in json.loads(response.json()):
                    role = i.get("role", "")
                    content = i.get("content", "")
                    print(f"```{role}\n{content.rstrip()}\n```")
            elif args.all.lower() == "json":
                print(response.json())
        else:
            print(json.loads(response.json())[-1]["content"])


if __name__ == '__main__':
    main()



"""
ToolMate AI Plugin - analyze web content

analyze web content with "AutoGen Retriever"

[TOOL_CALL]
"""

if not config.isLite:

    from toolmate import config
    from toolmate import print1, print2, print3, is_valid_url, downloadWebContent, ragRefineDocsPath, ragGetSplits, ragSearchContext, getRagPrompt
    from toolmate.utils.call_llm import CallLLM
    from toolmate.autoretriever import AutoGenRetriever

    def analyze_web_content(function_args):
        query = function_args.get("query") # required
        url = function_args.get("url") # required
        if not url or not is_valid_url(url):
            print1(f"'{url}' is not a valid url" if url else "No url is provided!")
            return "[INVALID]"
        config.stopSpinning()
        kind, filename = downloadWebContent(url)
        if not filename:
            return "[INVALID]"
        elif kind == "image":
            # call function "analyze image" instead if it is an image
            function_args = {
                "query": query,
                "files": [filename],
            }
            print3("Running tool: 'analyze_images'")
            return config.toolFunctionMethods["analyze_images"](function_args)

        if config.rag_useAutoRetriever and not config.llmInterface in ("vertexai", "googleai"):
            # process with AutoGen Retriever
            print2("AutoGen Retriever launched!")
            last_message = AutoGenRetriever().getResponse(filename, query, True)
            config.currentMessages += last_message
            print2("AutoGen Retriever closed!")
        else:
            print2("Retriever utility launched!")

            retrievedContext = ragSearchContext(ragGetSplits(ragRefineDocsPath(filename)), query)

            formatted_prompt = getRagPrompt(query, retrievedContext)

            messages = config.currentMessages[:-1] + [{"role": "user", "content" : formatted_prompt}]

            completion = CallLLM.regularCall(messages)
            print2(config.divider)
            config.toolmate.streamCompletion(completion)
            print2(config.divider)

            print2("Retriever utility closed!")
        return ""

    functionSignature = {
        "examples": [
            "analyze this url",
            "retrieve website information",
            "summarize this webpage",
        ],
        "name": "analyze_web_content",
        "description": "retrieve information from a webpage if an url is provided",
        "parameters": {
            "type": "object",
            "properties": {
                "url": {
                    "type": "string",
                    "description": """Return the given url. Return an empty string '' if it is not given.""",
                },
                "query": {
                    "type": "string",
                    "description": "Question that I ask about the given url",
                },
            },
            "required": ["query", "url"],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=analyze_web_content)